from typing import List
import smtplib
import ssl

from magicapi import g


port, smtp_server = g.settings.email_port, g.settings.email_smtp_server
sender_email, sender_password = g.settings.sender_email, g.settings.sender_password
from magicapi.Decorators.background_tasks import run_in_background


def send_email(text: str, recipients: List[str], subject: str = None):
    if subject:
        text = f"Subject: {subject}\n\n{text}"
    context = ssl.create_default_context()
    with smtplib.SMTP_SSL(smtp_server, port, context=context) as server:
        server.login(sender_email, sender_password)
        for recipient in recipients:
            server.sendmail(sender_email, recipient, text)


@run_in_background
def send_email_in_background(text: str, recipients: List[str], subject: str = None):
    send_email(text=text, recipients=recipients, subject=subject)


if __name__ == "__main__":
    send_email(
        subject="what do you know about Mactard Jones?",
        text="nothing, I hope",
        recipients=["fernando@basement.social"],
        # recipients=["kellycup8@gmail.com"],
    )



import os
from pathlib import Path
import datetime

import time

from mangum import Mangum

from fastapi import FastAPI, Request, APIRouter
from fastapi.middleware.cors import CORSMiddleware

from .config import MagicSettings, config_firestore

from magicapi.Globals.G import g


def add_cors(new_app):
    new_app.add_middleware(
        CORSMiddleware,
        allow_origins=["*"],
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["*"],
    )


def add_sentry(new_app):
    from magicapi.Services.Sentry import SentryAsgiMiddleware, HAS_SENTRY_DSN

    if HAS_SENTRY_DSN:
        new_app.add_middleware(SentryAsgiMiddleware)


def add_tasks_and_process_time_middleware(new_app):
    from magicapi.Models.task_executor import save_tasks
    from magicapi.Models.NotFound import save_not_found_from_request

    @new_app.middleware("http")
    async def add_process_time_and_tasks(request: Request, call_next):
        start_time = time.time()
        g.request = request
        g.app = new_app
        print("url path", request.url.path, "***url", request.url)
        if "aws.context" in request.scope:
            context = request.scope["aws.context"]
            if hasattr(context, "serverless_sdk"):
                print("added context to endpoint!")
                context.serverless_sdk.set_endpoint(request.url.path)
        response = await call_next(request)
        if response.status_code in [404, 405] and g.settings.save_calls:
            await save_not_found_from_request(request)
        response.headers["X-Tasks-Time"] = str(save_tasks())  # queue tasks
        response.headers["X-Process-Time"] = str(time.time() - start_time)
        return response


def add_process_time_middleware(new_app):
    @new_app.middleware("http")
    async def add_process_time(request: Request, call_next):
        start_time = time.time()
        print("url path", request.url.path, "***url", request.url)
        if "aws.context" in request.scope:
            context = request.scope["aws.context"]
            if hasattr(context, "serverless_sdk"):
                print("added context to endpoint!")
                context.serverless_sdk.set_endpoint(request.url.path)
        response = await call_next(request)
        response.headers["X-Process-Time"] = str(time.time() - start_time)
        return response


def add_hello_world_testing_route(new_app):
    @new_app.get("/", tags=["boilerplate"])
    def read_root():
        print("hello world!")
        return {
            "Welcome": "!",
            # "cwd": Path.cwd(),
            # "dir": os.listdir(),
        }


def add_addons(new_app, only_server: bool = False):
    add_cors(new_app)
    add_sentry(new_app)
    if not only_server:
        add_tasks_and_process_time_middleware(new_app)
    else:
        add_process_time_middleware(new_app)
    add_hello_world_testing_route(new_app)


def add_background_tasks():
    from magicapi.Decorators.background_tasks import run_in_background


def make_magic_router():
    from magicapi.RouteClasses import MagicCallLogger

    magic_router = APIRouter(route_class=MagicCallLogger)
    g.magic_router = magic_router


def add_magic_routers():
    g.app.include_router(g.magic_router)


def set_settings(settings=MagicSettings()):
    g.settings = settings


def create_app(config_settings=MagicSettings(), only_server: bool = False):
    set_settings(config_settings)
    print("creating_app", datetime.datetime.now())
    new_app = FastAPI(
        title=config_settings.app_name,
        version=config_settings.version,
        root_path="" if config_settings.local else f"/{config_settings.stage}",
    )

    g.app = new_app
    if not only_server:
        add_background_tasks()
    if not only_server:
        make_magic_router()

    add_addons(g.app, only_server)

    config_firestore(g.settings.service_account_name)

    return g.app


from fastapi.routing import APIRoute


def use_route_names_as_operation_ids(app: FastAPI) -> None:
    for route in app.routes:
        if isinstance(route, APIRoute):
            route.operation_id = route.name


def create_handler(app):
    def handler(event, context):
        """Handler to give to lambda... which wraps FastAPI"""
        if event.get("source") in ["aws.events", "serverless-plugin-warmup"]:
            print("Lambda is warm, now!!!")
            if hasattr(context, "serverless_sdk"):
                context.serverless_sdk.set_endpoint("/warmup")
            return {}

        asgi_handler = Mangum(app, api_gateway_base_path=g.settings.stage)
        response = asgi_handler(event, context)
        return response

    g.handler = handler
    return handler



from os.path import dirname, basename, isfile, join, exists
import glob


def get_all_subfiles_and_dirs(file_name):

    # first get all of the python files
    modules = glob.glob(join(dirname(file_name), "*.py"))
    __all__v = [
        basename(f)[:-3] for f in modules if isfile(f) and not f.endswith("__init__.py")
    ]

    # now import all of the directories
    init = "__init__.py"
    dirs_with_init = glob.glob(join(dirname(file_name), f"*/{init}"))
    for dir_with_init in dirs_with_init:
        d = dir_with_init.replace(f"/{init}", "")
        __all__v.append(basename(d))

    return __all__v



from typing import List
import requests

from magicapi import g


def send_email(text: str, recipients: List[str], subject: str = "", reply_to: str = None):
    data = {
        "from": f"{g.settings.mailgun_sender_name} <{g.settings.mailgun_sender_email}>",
        "to": recipients,
        "subject": subject,
        "text": text,
    }
    if "html" in data["text"]:
        data["html"] = data["text"]
        del data["text"]
    if reply_to:
        data['h:Reply-To'] = reply_to
    return requests.post(
        f"https://api.mailgun.net/v3/{g.settings.mailgun_domain_name}/messages",
        auth=("api", g.settings.mailgun_private_api_key),
        data=data,
    )


if __name__ == "__main__":
    resp = send_email(
        recipients=["kellycup8@aim.com"], text="jey man", subject="yessir"
    )
    print("resp content", resp.content)



import sys
import os
from pathlib import Path
import shutil
import subprocess

import requests
import typer

app = typer.Typer()

VENV_NAME = "magic_venv"

REPO_NAME = "jerber/MagicAPI-Boilerplate"

MAIN_FILENAME = "main.py"

START_FILENAME = "start.py"


@app.command()
def sowell():
    """
    Load the portal gun
    """

    typer.echo("IS GOAT")


def download_boilerplate_git_folder(output_filename):
    github_repo_url = f"https://api.github.com/repos/{REPO_NAME}/tarball/master"
    r = requests.get(github_repo_url, stream=True)
    with open(output_filename, "wb") as f:
        for chunk in r.raw.stream(1024, decode_content=False):
            if chunk:
                f.write(chunk)


def make_venv(venv_name: str = VENV_NAME):
    os.system(f"python3 -m venv {venv_name}")
    os.system(f"source {venv_name}/bin/activate && pip install -r requirements.txt")


@app.command()
def create(project_name: str = typer.Argument("magic-server"), replace: bool = False):
    # TODO add some printing of status
    project_path = Path(project_name)
    if project_path.exists() and not replace:
        typer.secho(
            "There is already a project that exists with this name. If you would like to replace this project, "
            "add the flag --replace to your command",
            fg=typer.colors.MAGENTA,
        )
        raise typer.Exit()

    zip_filename = "boilerplate.tar.gz"
    download_boilerplate_git_folder(zip_filename)

    temp_dir = Path("._magic_")
    shutil.unpack_archive(zip_filename, temp_dir)
    git_folder = list(temp_dir.glob("*"))[0]

    if project_path.exists():
        shutil.rmtree(project_path)

    git_folder.replace(project_path)

    shutil.rmtree(temp_dir)
    os.remove(zip_filename)

    typer.secho(
        f"Your app {project_name} has been created 🚀. Run 'cd {project_name}' and then "
        f"'magic dev' to start your first magic server 🎩!",
        fg=typer.colors.BRIGHT_GREEN,
    )


@app.command()
def dev(venv_name: str = typer.Argument(VENV_NAME), create_venv: bool = True):
    main_filename = "main.py"
    if not Path(main_filename).exists():
        typer.secho(
            "Cannot find the main.py file. Are you sure you created this app with magic?",
            fg=typer.colors.MAGENTA,
        )
        raise typer.Exit()

    command = f"export LOCAL=1 && python {main_filename}"

    out = subprocess.run(["which", "python"], stdout=subprocess.PIPE)
    interpreter_path = out.stdout.decode("utf-8").strip()

    # if they are using a venv, let them use it, but maybe install everything on it?
    if f"{venv_name}/bin/python" not in interpreter_path and create_venv:
        typer.echo(
            "No venv is active, will check to see if one exists and make one if not."
        )
        venv_path = Path(venv_name)
        if not venv_path.exists():
            typer.echo(f"No venv detected, will make one now with name {venv_name}")
            make_venv(venv_name)
        command = f"source {venv_name}/bin/activate && {command}"

    typer.secho(
        "To get interactive docs, go to http://0.0.0.0:8000/docs 🎩!",
        fg=typer.colors.BRIGHT_GREEN,
    )
    os.system(command)


@app.command()
def start():
    typer.secho(
        "To get interactive docs, go to http://0.0.0.0:8000/docs 🎩!",
        fg=typer.colors.BRIGHT_GREEN,
    )
    command = f"python {START_FILENAME}"
    os.system(command)


@app.command()
def test(url: str = typer.Argument("")):
    os.environ['testing_url'] = url
    command = f"pytest -s -v"
    os.system(command)


@app.command()
def deploy():
    command = f"sls deploy"
    os.system(command)


@app.command()
def deploy_again():
    command = f"sls deploy --function app"
    os.system(command)


@app.command()
def version():
    typer.echo("9")


if __name__ == "__main__":
    app()



from functools import wraps

# from magicapi import settings
from magicapi import g
from magicapi.Errors import BackendException

from magicapi.Services.Doorman import CurrentUser

from magicapi.Services import Segment

from magicapi.Decorators.helpers import async_safe


def segment(keywords=None):
    def inner_function(f):
        @wraps(f)
        async def wrapper(*args, **kwargs):
            if not g.settings.segment_write_key:
                raise BackendException(
                    message="You cannot use the segment decorator without the SEGMENT_WRITE_KEY env variable."
                )
            for key, val in kwargs.items():
                if issubclass(val.__class__, CurrentUser):
                    uid = val.uid
                    action = f.__name__
                    segment_d = {}
                    if keywords:
                        segment_d = {k: kwargs.get(k) for k in keywords}
                    Segment.track(uid, action, segment_d)
            return await async_safe(f, *args, **kwargs)

        return wrapper

    return inner_function



from datetime import datetime

from fastapi import Request, Response

from fastapi.routing import APIRoute
from typing import Callable

from magicapi.Models.Call import make_call_from_request_and_response
from magicapi.Models.task_executor import save_tasks

from magicapi.Services.Sentry import HAS_SENTRY_DSN, capture_exception

from magicapi.Errors import IgnoreException


class MagicCallLogger(APIRoute):
    def get_route_handler(self) -> Callable:
        original_route_handler = super().get_route_handler()

        async def custom_route_handler(request: Request) -> Response:
            response = None
            error = None
            time_received = datetime.utcnow()
            try:
                response: Response = await original_route_handler(request)
            # except IgnoreException as e:
            #     raise e
            except Exception as e:
                if HAS_SENTRY_DSN:
                    capture_exception(e)
                if "aws.context" in request.scope:
                    context = request.scope["aws.context"]
                    if hasattr(context, "serverless_sdk"):
                        print("Saving sls context for error!")
                        context.serverless_sdk.capture_exception(e)
                print("Error in call route", e.__dict__)
                error = e
            time_done = datetime.utcnow()
            delta = time_done - time_received
            secs_took = delta.seconds + delta.microseconds / 1_000_000
            times = {
                "time_received": time_received,
                "time_done": time_done,
                "secs_took": secs_took,
            }
            await make_call_from_request_and_response(request, response, error, times)
            if error:
                try:
                    save_tasks()
                except Exception as save_task_error:
                    print("[ERROR] save task error", save_task_error)
                raise error
            return response

        return custom_route_handler



from typing import Literal
from pathlib import Path
import os
from pydantic import BaseSettings


def config_firestore(service_account_name):
    service_account_path = find_file(service_account_name)
    if service_account_path:
        # will this work on windows?
        os.environ["GOOGLE_APPLICATION_CREDENTIALS"] = os.path.join(
            *service_account_path.parts
        )

        os.environ["HasDB"] = "1"

        # create the firebase connection with the service account
        from magicdb import db

        db.conn


def find_file(glob_string: str, levels: int = 10):
    d = Path.cwd()
    for _ in range(levels):
        envs = list(d.glob(glob_string))
        if envs:
            return envs[0]
        d = d.parent


class MagicSettings(BaseSettings):
    app_name: str = "GOAT Server"
    version: str = "0.0.1"

    service: str = "default_service"

    # for span sls
    use_span: bool = True

    # for dev
    print_level = 1

    # will be loaded first
    local: bool = False

    # for firestore
    service_account_name = "my-service-account.json"

    # for doorman
    doorman_public_project_id: str = None
    firebase_project_id: str = None
    cloud_function_location: str = None

    # for twilio
    twilio_account_sid: str = None
    twilio_auth_token: str = None
    twilio_messaging_service_sid: str = None
    twilio_status_callback: str = None
    from_number: str = None

    # for segment
    segment_write_key: str = None

    # for dynamo
    tasks_table_name: str = None

    # for routing
    stage: str = "dev"

    # for saving calls
    save_calls: bool = True

    # for email
    email_port: int = 465
    email_smtp_server: str = "smtp.gmail.com"
    sender_email: str = None
    sender_password: str = None

    # for mailgun
    mailgun_private_api_key: str = None
    mailgun_domain_name: str = None
    mailgun_sender_name: str = None
    mailgun_sender_email: str = None

    # for magic link
    company_name: str = None

    # for stripe
    stripe_api_key_sandbox: str = None
    stripe_api_key_production: str = None
    stripe_env: Literal["sandbox", "production"] = None

    # for sentry
    sentry_dsn: str = None
    sentry_traces_sample_rate: float = 0.03

    # for background save call url
    background_tasks_url: str = None

    class Config:
        env_file = find_file(glob_string="*.env") or ".env"


# If you want to make the env variables from .env available
# from dotenv import load_dotenv
# if os.path.exists(env_path):
#     load_dotenv(env_path)

# settings = MagicSettings()

# config_firestore(settings.service_account_name)



from functools import wraps
from pydantic import BaseModel
import inspect
from magicapi import g

from magicapi.Decorators.helpers import async_safe


def parse_objects(f):
    @wraps(f)
    async def wrapper(*args, **kwargs):
        sigs = inspect.signature(f)
        parameters = sigs.parameters
        args = list(args)
        for i, (arg, param) in enumerate(zip(args.copy(), parameters.values())):
            this_class = param.annotation
            if inspect.isclass(this_class) and issubclass(param.annotation, BaseModel):
                args[i] = this_class.parse_obj(arg)

        for var_name, var_val in kwargs.copy().items():
            this_class = parameters.get(var_name).annotation

            if (
                var_name in parameters
                and inspect.isclass(this_class)
                and issubclass(this_class, BaseModel)
            ):
                kwargs[var_name] = this_class.parse_obj(var_val)

        if g.settings.print_level > 1:
            print("PARSE OBJS", "args", args, "kwargs", kwargs)

        return await async_safe(f, *args, **kwargs)

    return wrapper



import sentry_sdk
from sentry_sdk import capture_exception
from sentry_sdk.integrations.asgi import SentryAsgiMiddleware

from magicapi import g

from magicapi.Decorators.background_tasks import run_in_background

HAS_SENTRY_DSN = g.settings.sentry_dsn
if HAS_SENTRY_DSN:
    sentry_sdk.init(
        dsn=g.settings.sentry_dsn,
        traces_sample_rate=g.settings.sentry_traces_sample_rate,
        send_default_pii=True,
        request_bodies="always",
    )



from datetime import datetime
from magicdb import FrontendParserModel, MagicModel


class Params(FrontendParserModel):
    args: list
    kwargs: dict


class BackgroundTask(MagicModel):
    params: Params
    created_at: datetime
    finished_at: datetime = None
    task_took: datetime = None



from fastapi import Request as FastAPIRequest

from magicdb.Models import MagicModel, DateModel
from magicapi.Models.Call import Request, make_request_obj_from_request

from magicapi.Decorators.background_tasks import run_in_background
from magicapi.Decorators.parse_objects import parse_objects


class NotFound(DateModel):
    request: Request

    class Meta:
        collection_name = '_not_found'


async def save_not_found_from_request(req: FastAPIRequest):
    req_obj = await make_request_obj_from_request(req)
    nf = NotFound(request=req_obj)
    save_not_found(nf)


@run_in_background
@parse_objects
def save_not_found(not_found: NotFound):
    not_found.save()



import boto3
from datetime import datetime
from magicdb import FrontendParserModel, DateModel, MagicModel

# from magicapi import settings
# from magicapi.Globals.G import g


from magicapi import g

dynamodb = boto3.resource("dynamodb")
table = dynamodb.Table(g.settings.tasks_table_name)

# this is deprecated, only used if you want to use dynamo as Q


class TaskParams(MagicModel):
    args: list
    kwargs: dict


# for now cannot accept an object that does not have the dict ability...
class Task(DateModel):
    task_id: str
    url: str
    status: str
    sent: bool
    secret_token: str
    params: str = None
    local: bool

    class Meta:
        collection_name = "_tasks"

    @staticmethod
    def get_table():
        return table



import time
from magicapi import settings

from magicapi.Decorators.helpers import async_safe


def timeit(method):
    async def timed(*args, **kwargs):
        s = time.time()
        result = await async_safe(method, *args, **kwargs)
        e = time.time()
        if settings.print_level > 0:
            print(f"{method.__name__} took {(e - s)*1_000} ms")
        return result

    return timed



from typing import Callable
from magicdb import auth
from magicapi.Services.MagicLink.template import make_template_and_subject


def make_magic_link(email_address: str, redirect_url: str):
    action_code_settings = auth.ActionCodeSettings(url=redirect_url)
    magic_link = auth.generate_sign_in_with_email_link(
        email_address, action_code_settings
    )
    return magic_link


def get_magic_link_template_and_subject(email_address: str, redirect_url: str):
    magic_link = make_magic_link(email_address=email_address, redirect_url=redirect_url)
    magic_template, subject = make_template_and_subject(
        email_address=email_address, magic_link=magic_link
    )
    return magic_template, subject


def send_magic_link_email(
    sender_function: Callable, email_address: str, redirect_url: str
):
    """The sender function must take in a text, subject, and recipients"""
    magic_template, subject = get_magic_link_template_and_subject(
        email_address=email_address, redirect_url=redirect_url
    )
    return sender_function(
        text=magic_template, subject=subject, recipients=[email_address]
    )


if __name__ == "__main__":
    magic_template, subject = make_template_and_subject(
        email_address="kellycup8@gmail.com", magic_link="www.penguincbd.com"
    )
    print(magic_template, subject)



import time
import requests
import os
from functools import wraps
import functools
from fastapi import Request, Body

from magicapi.Models.Task import Task, TaskParams
from magicapi.Models.BackgroundTask import BackgroundTask, Params

import inspect
from fastapi import HTTPException, APIRouter
import json
from datetime import datetime

from magicapi.Utils.random_utils import random_str
from magicapi.Errors import BackendException

from magicapi import g

from magicapi.Decorators.helpers import async_safe


def run_in_background_no_dynamo(f=None, background_url: str = None):
    if not f:
        if background_url and g.settings.local:
            background_url = None
        return functools.partial(
            run_in_background_no_dynamo, background_url=background_url
        )

    router_path = f"/run_in_background_no_dynamo/{f.__name__}"
    new_r = APIRouter()

    @new_r.post(router_path, tags=["background_tasks"], include_in_schema=False)
    async def endpoint(task: BackgroundTask):
        start = time.time()
        await async_safe(f, *task.params.args, **task.params.kwargs)
        time_took = time.time() - start
        task.finished_at = datetime.utcnow()
        task.task_took = datetime.utcnow()
        print(f"Finished {f.__name__}, task {task.id} in {time_took}.")
        return task

    g.app.include_router(new_r)

    @wraps(f)
    def wrapper(*args, **kwargs):
        task = BackgroundTask(
            params=Params(args=list(args), kwargs=kwargs), created_at=datetime.utcnow()
        )
        headers = {"Content-Type": "application/json", "Accept": "application/json"}
        url = f"{background_url or g.base_url}{router_path}"
        try:
            requests.post(
                url=url,
                data=task.json(),
                headers=headers,
                timeout=0.01,
            )
        except requests.exceptions.ReadTimeout:
            print(f"Sent {f.__name__} background task {task.id}, {url=}")

    return wrapper


def run_in_background(f=None, background_url: str = None):
    if not f:
        if background_url and g.settings.local:
            background_url = None
        return functools.partial(run_in_background, background_url=background_url)
    router_path = f"/run_in_background/{f.__name__}"
    new_r = APIRouter()

    @new_r.post(router_path, tags=["background_tasks_dynamo"], include_in_schema=False)
    async def endpoint(
        request: Request,
        task_id: str = Body(...),
        secret_token: str = Body(...),
        params: str = Body(...),
    ):
        if g.settings.print_level > 1:
            print("endpoint just received!", "params", params)
        # get the task from dynamo
        response = Task.get_table().get_item(Key={"task_id": task_id})
        task_dict = response.get("Item")
        if not task_dict:
            raise HTTPException(status_code=404, detail="Invalid task id.")

        task = Task(**task_dict)
        if not task or secret_token != task.secret_token:
            raise HTTPException(status_code=404, detail="Invalid task request.")

        if task.status != "in-lambda":
            # when local sometimes this happens before lambda changes queued to in-lambda
            if not task.local:
                raise HTTPException(
                    status_code=404, detail="This task was already completed."
                )

        j_params = json.loads(params)
        args = j_params.get("args", [])
        kwargs = j_params.get("kwargs", {})
        if g.settings.print_level > 1:
            print("inspect", inspect.signature(f), "a", args, "k", kwargs)

        await async_safe(f, *args, **kwargs)

        return {"success": True, "message": ""}

    g.app.include_router(new_r)

    @wraps(f)
    def wrapper(*args, **kwargs):
        if not os.getenv("HasDB"):
            raise BackendException(
                message="You cannot add tasks if you do not add a DB!"
            )

        if g.settings.print_level > 1:
            print("given to function", "args", args, "kwargs", kwargs)
        task_params = TaskParams(args=list(args), kwargs=kwargs)
        now = datetime.utcnow()

        base_url = background_url or g.base_url
        task = Task(
            task_id=random_str(30),
            url=base_url + router_path,
            status="queued",
            sent=False,
            secret_token=random_str(50),
            created_at=now,
            last_updated=now,
            params=task_params.json(),
            local=g.settings.local,
        )
        g.tasks.append(task)

        return task.task_id

    return wrapper



from magicdb.Models import FrontendParserModel
from pydantic import EmailStr
from firebase_admin import auth


class CurrentUser(FrontendParserModel):
    uid: str
    phone_number: str = None
    email: EmailStr = None
    email_verified: bool = None
    auth_time: int
    iat: int
    exp: int
    iss: str = None
    aud: str = None
    user_id: str = None
    sub: str = None
    firebase: dict = {}
    display_name: str = None

    def update(self, **kwargs):
        return auth.update_user(self.uid, **kwargs)



import analytics

from magicapi import g

from magicapi.Decorators.background_tasks import run_in_background

analytics.write_key = g.settings.segment_write_key
analytics.sync_mode = True


@run_in_background
def identify(*args, **kwargs):
    return analytics.identify(*args, **kwargs)


@run_in_background
def track(*args, **kwargs):
    return analytics.track(*args, **kwargs)


@run_in_background
def page(*args, **kwargs):
    return analytics.page(*args, **kwargs)


@run_in_background
def screen(*args, **kwargs):
    return analytics.screen(*args, **kwargs)


@run_in_background
def group(*args, **kwargs):
    return analytics.group(*args, **kwargs)


@run_in_background
def alias(*args, **kwargs):
    return analytics.alias(*args, **kwargs)



import phonenumbers
from pydantic import BaseModel
import re


class PhoneNumber(str):
    @classmethod
    def __get_validators__(cls):
        # one or more validators may be yielded which will be called in the
        # order to validate the input, each validator will receive as an input
        # the value returned from the previous validator
        yield cls.validate

    @classmethod
    def __modify_schema__(cls, field_schema):
        # __modify_schema__ should mutate the dict it receives in place,
        # the returned value will be ignored
        field_schema.update(
            # some example phonenumbers
            examples=["609-356-9384", "+16093284938"],
        )

    @classmethod
    def validate(cls, v):
        if not isinstance(v, str):
            raise TypeError("string required")

        # two edge cases for american numbers-- all non american numbers must have a +<countrycode>
        # if there is no + and the len of all digits is 10 and there is no leading 1... add a +1
        #   if there len of all digits is 11 and there is a leading 1, add a +

        if "+" not in v:
            digits = "".join(re.findall(r"\d+", v))
            if len(digits) == 10 and digits[0] != "1":
                v = "+1" + v
            elif len(digits) == 11 and digits[0] == "1":
                v = "+" + v

        try:
            x = phonenumbers.parse(v)
            formatted = phonenumbers.format_number(
                x, phonenumbers.PhoneNumberFormat.E164
            )
            return cls(formatted)
        except phonenumbers.phonenumberutil.NumberParseException as err:
            raise ValueError(err, "given phone number value:", v)

    def __repr__(self):
        return f"PhoneNumber({super().__repr__()})"


if __name__ == "__main__":

    class Model(BaseModel):
        phone_number: PhoneNumber

    model = Model(phone_number="(4555)555-5555")
    print("model", model)
    print("model phone num", model.phone_number)
    print(type(model.phone_number))



from twilio.rest import Client
import os
from functools import wraps
from magicapi import g

from magicapi.Errors import TwilioException

from twilio.base.exceptions import TwilioRestException

try:
    client = Client(g.settings.twilio_account_sid, g.settings.twilio_auth_token)
except Exception as e:
    print("Twilio SID or AUTH TOKEN not given.")

FROM_NUMBER = os.environ.get("FROM_NUMBER")


def need_twilio_vars(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        twilio_vars = [g.settings.twilio_account_sid, g.settings.twilio_auth_token]
        if None in twilio_vars:
            raise TwilioException(
                message="Not all Twilio credentials found in .env file. Need TWILIO_ACCOUNT_SID and TWILIO_AUTH_TOKEN."
            )
        return f(*args, **kwargs)

    return wrapper


def need_copilot_vars(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        twilio_vars = [g.settings.twilio_messaging_service_sid]
        if None in twilio_vars:
            raise TwilioException(
                message="Not all Twilio co-pilot credentials found in .env file. Need MESSAGING_SERVICE_SID."
            )
        return f(*args, **kwargs)

    return wrapper


@need_twilio_vars
def send_text(to, body, from_number=FROM_NUMBER):
    message = client.messages.create(
        body=body,
        from_=from_number,
        to=to,
    )
    return message.sid


@need_twilio_vars
def send_text_with_copilot(to, body, status_callback=g.settings.twilio_status_callback):
    try:
        message = client.messages.create(
            body=body,
            messaging_service_sid=g.settings.twilio_messaging_service_sid,
            to=to,
            status_callback=status_callback,
        )
        return message.sid
    except TwilioRestException as error:
        print("Twilio error here", error)
        print("*THIS WAS THE ERRORED MESSAGE*", body)



import sys
from magicdb.Models import DateModel
from datetime import datetime
from typing import Any
from pydantic import BaseModel
from fastapi import Request as FastAPIRequest
from fastapi import Response as FastApiResponse
import json
from typing import Optional

from magicapi.Utils.random_utils import random_str

from magicapi.Decorators.background_tasks import run_in_background
from magicapi.Decorators.parse_objects import parse_objects

# from magicapi import settings
from magicapi import g


class Error(BaseModel):
    error_class: str
    error_dict: Any


class Response(BaseModel):
    body: Any
    headers: dict
    status_code: int


class Request(BaseModel):
    request_id: str
    body: Any
    headers: dict
    cookies: dict
    url: str
    url_path: str
    root_url: str
    query_params: dict
    ip_address: str
    method: str
    scheme: str
    port: Optional[int]
    # added from aws context
    aws_request_id: str = None
    log_group_name: str = None
    log_stream_name: str = None
    function_name: str = None
    memory_limit_in_mb: str = None
    # added from aws event
    request_id: str = None
    domain_name: str = None
    api_id: str = None
    # created_urls
    log_group_url: str = None
    stream_url: str = None


class Times(BaseModel):
    time_received: datetime
    time_done: datetime
    secs_took: float
    get_size_time: float = None


class Call(DateModel):
    request: Request
    response: Response = None
    error: Error = None
    times: Times
    message: str = None

    class Meta:
        collection_name = "_calls"


@run_in_background(background_url=g.settings.background_tasks_url)
@parse_objects
def save_call(call: Call):
    try:
        call.save()
    except Exception as error:
        if "400 Cannot convert an array value in an array value." in str(error):
            print("In save call, google error", error)
            call.request.body = str(call.request.body)
            call.response.body = str(call.response.body)
            call.save()
            return
        else:
            raise error


async def get_request_body(request: FastAPIRequest):
    # first try json, then body, then form...
    errors = []
    try:
        result = await request.json() or None
        return result
    except Exception as json_error:
        errors.append(f"json_error {json_error}")

    try:
        result = await request.form() or None
        return dict(result)
    except Exception as form_error:
        errors.append(f"form_error {form_error}")

    try:
        result = await request.body() or None
        return result
    except Exception as body_error:
        errors.append(f"body_error {body_error}")

    if g.settings.print_level > 1:
        print("get_request_body_errors", errors)

    return None


def make_body_jsonable(body):
    try:
        json.dumps(body)
        return body
    except TypeError as _:
        pass

    # now try making string
    try:
        body_str = str(body)
        return body_str
    except TypeError as _:
        pass

    return None


def safe_loads(body):
    if body is None:
        return body
    try:
        return json.loads(body)
    except TypeError as _:
        pass

    try:
        return str(body)
    except TypeError as _:
        pass

    return None


async def make_request_obj_from_request(request: FastAPIRequest):
    body = await get_request_body(request)
    jsonable_body = make_body_jsonable(body)

    request_obj = Request(
        request_id=random_str(30),
        body=jsonable_body,
        headers=dict(request.headers),
        cookies=dict(request.cookies),
        url=str(request.url),
        url_path=request.url.path,
        root_url=str(request.url).replace(request.url.path, ""),
        query_params=dict(request.query_params),
        ip_address=request.client.host,
        method=request.method,
        scheme=request.url.scheme,
        port=request.url.port,
    )

    aws_fields = [
        "aws_request_id",
        "log_group_name",
        "log_stream_name",
        "function_name",
        "memory_limit_in_mb",
    ]

    if "aws.context" in request.scope:
        context = request.scope["aws.context"]
        [setattr(request_obj, field, getattr(context, field)) for field in aws_fields]

        cloudwatch_base_url = "https://console.aws.amazon.com/cloudwatch/home?region=us-east-1#logStream:group="
        request_obj.log_group_url = f"{cloudwatch_base_url}{request_obj.log_group_name}"
        request_obj.stream_url = (
            f"{request_obj.log_group_url};stream={request_obj.log_stream_name}".replace(
                "[$LATEST]", "%5B$LATEST%5D"
            ).replace("logStream", "logEventViewer")
        )

    if "aws.event" in request.scope:
        event = request.scope["aws.event"]
        request_context = event.get("requestContext")
        if request_context:
            request_obj.request_id = request_context.get("requestId")
            request_obj.domain_name = request_context.get("domainName")
            request_obj.api_id = request_context.get("apiId")

    return request_obj


async def make_call_from_request_and_response(
    request: FastAPIRequest,
    response: Optional[FastApiResponse],
    error: Optional[Exception],
    times_dict: dict,
):
    if not g.settings.save_calls:
        return

    request_obj = await make_request_obj_from_request(request)

    response_obj = (
        None
        if not response or not hasattr(response, "body")
        else Response(
            body=safe_loads(response.body),
            headers=dict(response.headers),
            status_code=response.status_code,
        )
    )

    error_obj = (
        None
        if not error
        else Error(
            error_class=str(error.__class__),
            error_dict=make_body_jsonable(error.__dict__),
        )
    )

    # maybe wrap this whole thing w a try catch just in case so it does not fuck up everything else
    # will do this once I test this out for a while... prob

    call = Call(
        request=request_obj,
        response=response_obj,
        error=error_obj,
        times=Times(**times_dict),
    )

    DYNAMO_DB_MAX_SIZE = 400_000

    def size_under(obj, seen=None, max_size=DYNAMO_DB_MAX_SIZE):
        """Recursively finds size of objects"""
        size = sys.getsizeof(obj)
        if seen is None:
            seen = set()
        obj_id = id(obj)
        if obj_id in seen:
            return 0
        # Important mark as seen *before* entering recursion to gracefully handle
        # self-referential objects
        seen.add(obj_id)
        if isinstance(obj, dict):
            size += sum([size_under(v, seen) for v in obj.values()])
            size += sum([size_under(k, seen) for k in obj.keys()])
            if size > max_size:
                return size
        elif hasattr(obj, "__dict__"):
            size += size_under(obj.__dict__, seen)
            if size > max_size:
                return size
        elif hasattr(obj, "__iter__") and not isinstance(obj, (str, bytes, bytearray)):
            size += sum([size_under(i, seen) for i in obj])
            if size > max_size:
                return size
        return size

    import time

    start_size_time = time.time()
    size_of_call = size_under(call)
    size_time_took = time.time() - start_size_time
    call.times.get_size_time = size_time_took

    if size_of_call > DYNAMO_DB_MAX_SIZE:
        print("size of call was", size_of_call)
        if call.request:
            call.request.body = str(call.request.body)[0:300]
        if call.response:
            call.response.body = str(call.response.body)[0:300]
        if call.error:
            call.error.error_dict = str(call.error.error_dict)[0:300]
        call.message = f"Original size {size_of_call} was too big so cut out request and response body."
    save_call(call)



from magicapi import g

# email_body = f'<html><a href="<LINK>" target=“_blank”>Click here to sign in to {g.settings.company_name}.</a></html>'

email_body = '''
<html>
  Hello,
  <p>
    We received a request to sign in to <COMPANY_NAME> using this email address. If you want to sign in with your <EMAIL_ADDRESS> account, click this link:
  </p>
  <p>
    <a
    href="<LINK>"
    target="“_blank”"
    >Click here to sign in to <COMPANY_NAME>.</a
  >
  </p>
  <p>
      If you did not request this link, you can safely ignore this email.
  </p>
  <p>
      Thanks,
  </p>
  <p>
      Your <COMPANY_NAME> team
  </p>
</html>
'''

subject = f"Sign in to {g.settings.company_name}"


def make_template_and_subject(email_address: str, magic_link: str):
    new_body = email_body.replace("<LINK>", magic_link).replace(
        "<EMAIL_ADDRESS>", email_address
    ).replace('<COMPANY_NAME>', g.settings.company_name)
    return new_body, subject



from fastapi import FastAPI, Request
from magicapi.config import MagicSettings


class G:
    def __init__(self, app: FastAPI = None, request: Request = None, settings: MagicSettings = None) -> None:
        self._app = app
        self._request = request
        self.tasks = []
        self.settings = settings

    @property
    def app(self) -> FastAPI:
        return self._app

    @app.setter
    def app(self, app: FastAPI) -> None:
        self._app = app

    @property
    def request(self) -> Request:
        return self._request

    @request.setter
    def request(self, request: Request) -> None:
        self._request = request

    @property
    def base_url(self) -> str:
        url = str(self.request.url)
        path = self.request.url.path
        url = url[: url.rindex(path)]
        url = url if self.settings.local else url + f"/{self.settings.stage}"
        return url

    @property
    def url(self) -> str:
        return str(self.request.url)


g = G()



from fastapi import Request

from magicapi import g

from fastapi.responses import JSONResponse

from pydantic import BaseModel


class MagicException(Exception):
    def __init__(self, message: str, json_response: dict = None):
        print(f"*MagicException*: {message=}")
        self.status_code: int = 452
        self.message: str = message
        self.json_response: dict = json_response or {}

    class MagicExceptionModel(BaseModel):
        status_code: int = 452
        message: str
        json_response: dict = {}

    def to_model(self) -> MagicExceptionModel:
        if self.message and not type(self.message) == str:
            self.message = str(self.message)
        return self.MagicExceptionModel(**self.__dict__)

    @classmethod
    def get_model(cls) -> type(MagicExceptionModel):
        return cls.MagicExceptionModel


class BackendException(MagicException):
    pass


class FrontendException(MagicException):
    pass


class FirestoreException(MagicException):
    pass


class TwilioException(MagicException):
    pass


class IgnoreException(MagicException):
    pass


@g.app.exception_handler(MagicException)
def backend_exception_handler(request: Request, exc: MagicException):
    print("EXC", exc.__dict__)
    magic_exception_model = exc.to_model()
    return JSONResponse(
        status_code=magic_exception_model.status_code,
        content=magic_exception_model.dict(),
    )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['magicapi',
 'magicapi.Decorators',
 'magicapi.Errors',
 'magicapi.FieldTypes',
 'magicapi.Globals',
 'magicapi.Models',
 'magicapi.RouteClasses',
 'magicapi.Services',
 'magicapi.Services.Doorman',
 'magicapi.Services.Email',
 'magicapi.Services.MagicLink',
 'magicapi.Services.Mailgun',
 'magicapi.Services.Segment',
 'magicapi.Services.Sentry',
 'magicapi.Services.Stripe',
 'magicapi.Services.Twilio',
 'magicapi.Utils']

package_data = \
{'': ['*']}

install_requires = \
['analytics-python>=1.2.9,<2.0.0',
 'boto3>=1.14.23,<2.0.0',
 'fastapi[all]>=0.61.0,<0.62.0',
 'magicdb>=0.2.94,<0.3.0',
 'mangum>=0.9.2,<0.10.0',
 'phonenumbers>=8.12.6,<9.0.0',
 'pydantic[dotenv]>=1.7.1,<2.0.0',
 'requests>=2.24.0,<3.0.0',
 'sentry-sdk>=0.19.4,<0.20.0',
 'stripe>=2.49.0,<3.0.0',
 'twilio>=6.44.0,<7.0.0',
 'typer>=0.3.1,<0.4.0']

entry_points = \
{'console_scripts': ['magic = magicapi.cli:app']}

setup_kwargs = {
    'name': 'magicapi',
    'version': '0.1.123',
    'description': '',
    'long_description': '# Magic API\n\nThe Magic API!\n\n1) incr the toml version\n2) poetry publish --build\n\n\nFor CLI:\n# Magic CLI (magicapi has this built in)\n\nThe best server maker.\n\nSteps:\n\n1) pip install magicapi (make sure the pip points to the correct python (3.6+))\n\n2) cd into the directory you want the server to be in\n\n// use the cli to create the app\n\n3) magic create < app name >\n\n4) cd < app name >\n\n// now make the virtual env and pip install the requirements\n5) python3 -m venv venv\n\n6) source venv/bin/activate\n\n7) `pip install -r requirements.txt`\n8) `pip install --upgrade magicdb magicapi`\n\n// files needed to run\n8) Create an .env file in the < app name > directory.\n\na) Use the ".example_env" as a template. Fill in the env variables for the services you will need to use.\n\nb) The SERVICE and TASKS_TABLE_NAME are required for the app.\n\n// add firestore\n1) add the firestore service account json to the < app name > and name it "my-service-account.json"\n\n// to deploy to aws lambda\n\n1) Create a user on AWS and sign in via the CLI (https://docs.aws.amazon.com/cli/latest/userguide/cli-configure-quickstart.html)\n \n2) npm i\n\n2) magic deploy\n\nThat should be it for the setup!\n\nTime to run it!\n\nTo start a local server: magic start\n\nTo test local server (while local server is running): magic test\n\nTo deploy for the first time or any time you change the serverless yaml: magic deploy\n\nTo deploy when you just edited the app code: magic deploy_again\n\nTo update magicapi:\npip install --upgrade magicapi\n',
    'author': 'Jeremy Berman',
    'author_email': 'jerber@sas.upenn.edu',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/jerber/MagicAPI',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



__version__ = "0.1.0"

"""Imports the Routes and the app and glues them all together"""
# from .config import settings
from magicapi.Globals.G import g

from magicapi.app_factory import (
    create_app,
    create_handler,
    add_magic_routers,
    use_route_names_as_operation_ids,
)



import time
import json
import threading
import requests

from fastapi.encoders import jsonable_encoder

from magicapi import g
from .Task import Task

from botocore.exceptions import ClientError


def run_tasks_locally(task):
    # make a dict but make sure dates are properly parsed
    d = json.loads(task.json())
    resp = requests.post(task.url, json=d)
    print("resp from local task", resp.content)


def save_tasks():
    if not g.tasks:
        return 0

    start = time.time()
    count = 0
    saved_tasks = []

    try:
        with Task.get_table().batch_writer() as batch:
            while len(g.tasks) and count < 25:
                task = g.tasks.pop(0)
                batch.put_item(Item=jsonable_encoder(task))
                saved_tasks.append(task)
                count += 1
    except ClientError as dynamo_error:
        print("Dynamo error likely so cannot save response:", dynamo_error)
        raise dynamo_error

    if g.settings.local:
        for task in saved_tasks:
            threading.Thread(target=run_tasks_locally, args=(task,)).start()
    return time.time() - start



from typing import Optional
import requests
from functools import wraps

from fastapi import Depends, APIRouter
from fastapi.security import OAuth2PasswordBearer, OAuth2PasswordRequestForm
from .CurrentUser import CurrentUser

from magicapi import g
import firebase_admin
from firebase_admin import auth

from .errors import DoormanAuthException
from magicapi.Decorators.firestore import need_firestore
from magicapi.FieldTypes import PhoneNumber

# from magicapi import settings
from magicapi import g
from magicapi.RouteClasses import MagicCallLogger

LOCATION, PROJECT_ID, DOORMAN_ID = (
    g.settings.cloud_function_location or "us-central1",
    g.settings.firebase_project_id,
    g.settings.doorman_public_project_id,
)

ID_TOKEN_ENDPOINT: str = (
    f"https://{LOCATION}-{PROJECT_ID}.cloudfunctions.net/getIdToken"
)
DOORMAN_BACKEND_ENDPOINT: str = (
    "https://sending-messages-for-doorman.herokuapp.com/phoneLogic"
)

doorman_prefix = "/doorman"
bare_token_path = f"{doorman_prefix}/token"
token_url = (
    bare_token_path if g.settings.local else f"/{g.settings.stage}{bare_token_path}"
)
oath2_scheme = OAuth2PasswordBearer(tokenUrl=token_url)
oath2_scheme_or_none = OAuth2PasswordBearer(tokenUrl=token_url, auto_error=False)

doorman_router = APIRouter(route_class=MagicCallLogger)


def need_doorman_vars(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        doorman_vars = [LOCATION, PROJECT_ID, DOORMAN_ID]
        if None in doorman_vars:
            raise DoormanAuthException(
                message="Not all Doorman credentials found in .env file. Need DOORMAN_PUBLIC_PROJECT_ID, "
                "FIREBASE_PROJECT_ID, and CLOUD_FUNCTION_LOCATION"
            )
        return f(*args, **kwargs)

    return wrapper


@doorman_router.post("/login_with_phone")
@need_doorman_vars
def login_with_phone(phone_number: PhoneNumber):
    body = {
        "action": "loginWithPhone",
        "phoneNumber": phone_number,
        "publicProjectId": DOORMAN_ID,
    }
    resp = requests.post(DOORMAN_BACKEND_ENDPOINT, json=body).json()
    return resp


def sign_in_with_magic_link(form_data):
    body = {"email": form_data.username, "magicLink": form_data.password}

    magic_link_id_token_endpoint = ID_TOKEN_ENDPOINT.replace(
        "getIdToken", "getIdTokenFromMagicLink"
    )

    id_resp = requests.post(magic_link_id_token_endpoint, json=body).json()
    id_token = id_resp.get("idToken")
    if not id_token:
        print(id_resp)
        raise DoormanAuthException(message=str(id_resp))

    return {"access_token": id_token, "token_type": "bearer"}


@doorman_router.post("/token")
@need_doorman_vars
def login(form_data: OAuth2PasswordRequestForm = Depends()):
    if "@" in form_data.username:
        return sign_in_with_magic_link(form_data)
    body = {
        "action": "verifyCode",
        "phoneNumber": form_data.username,
        "code": form_data.password,
        "publicProjectId": DOORMAN_ID,
    }
    resp = requests.post(DOORMAN_BACKEND_ENDPOINT, json=body).json()
    backend_token = resp.get("token")
    if not backend_token:
        print(resp)
        raise DoormanAuthException(message=str(resp))

    id_resp = requests.post(ID_TOKEN_ENDPOINT, json={"token": backend_token}).json()
    id_token = id_resp.get("idToken")
    if not id_token:
        print(id_resp)
        raise DoormanAuthException(message=str(id_resp))

    return {"access_token": id_token, "token_type": "bearer"}


def get_current_user_from_token(token: str):
    try:
        decoded = auth.verify_id_token(token)
        current_user = CurrentUser(**decoded)
        print(
            f"Current User: {current_user.uid}, name: {current_user.display_name}, phone_number: {current_user.phone_number}"
        )
        return current_user
    except firebase_admin._token_gen.ExpiredIdTokenError as e:
        print("DOORMAN ERROR", e)
        raise DoormanAuthException(message=e)


@need_firestore
def get_current_user(token: str = Depends(oath2_scheme)):
    return get_current_user_from_token(token=token)


@need_firestore
def get_current_user_or_none(token: Optional[str] = Depends(oath2_scheme_or_none)):
    if not token:
        return None
    current_user = get_current_user_from_token(token)
    return current_user


@need_firestore
def get_current_user_raw(token: str = Depends(oath2_scheme)):
    try:
        decoded = auth.verify_id_token(token)
        return decoded
    except firebase_admin._token_gen.ExpiredIdTokenError as e:
        raise DoormanAuthException(message=e)


GET_USER = Depends(get_current_user)
GET_USER_OR_NONE = Depends(get_current_user_or_none)

g.app.include_router(doorman_router, prefix=doorman_prefix, tags=["Doorman"])



import os
from functools import wraps
from magicapi.Errors import FirestoreException

from magicapi.Decorators.helpers import async_safe


def need_firestore(f):
    @wraps(f)
    async def wrapper(*args, **kwargs):
        if not os.getenv("GOOGLE_APPLICATION_CREDENTIALS"):
            raise FirestoreException(
                message="You must supply a firestore service account to use this endpoint!"
            )
        return await async_safe(f, *args, **kwargs)

    return wrapper
